from .client import Client
from .packet import Packet
from . import utils

from .__version__ import __author__, __title__, __description__
from .__version__ import __url__, __version__, __credits__, __license__